# See End Of File For Licensing

import inspect
import itertools
from weakref import ref
from types import MethodType

from .utils import Sentinel
from .base import Model, Control, view


__all__ = [
    'List',
    'Dict',
    'Set',
    'Object',
    'Oracle',
    'Undefined',
]


Undefined = Sentinel('Undefined')


class List(Model, list):
    """An MVC enabled ``list``."""

    _control_setitem = Control('__setitem__')
    _control_delitem = Control('__delitem__')
    _control_insert = Control('insert')
    _control_append = Control('append')
    _control_extend = Control('extend')
    _control_pop = Control('pop')
    _control_clear = Control('clear')
    _control_remove = Control('remove')
    _control_sort = Control('sort').before('_control_rearrangement')
    _control_reverse = Control('reverse').before('_control_rearrangement')

    @_control_setitem.before
    def _control_setitem(self, call, notify):
        index = call.args[0]
        try:
            old = self[index]
        except KeyError:
            old = Undefined
        return index, old

    @_control_setitem.after
    def _control_setitem(self, answer, notify):
        index, old = answer.before
        new = self[index]
        if new is not old:
            notify(index=index, old=old, new=new)

    @_control_delitem.before
    def _control_delitem(self, call, notify):
        index = call.args[0]
        return index, self[index:]

    @_control_delitem.after
    def _control_delitem(self, answer, notify):
        index, old = answer.before
        for i, x in enumerate(old):
            try:
                new = self[index + i]
            except IndexError:
                new = Undefined
            notify(index=(i + index), old=x, new=new)

    @_control_insert.before
    def _control_insert(self, call, notify):
        index = call.args[0]
        return index, self[index:]

    @_control_insert.after
    def _control_insert(self, answer, notify):
        index, old = answer.before
        for i in range(index, len(self)):
            try:
                o = old[i]
            except IndexError:
                o = Undefined
            notify(index=i, old=o, new=self[i])

    @_control_append.after
    def _control_append(self, answer, notify):
        notify(index=len(self) - 1, old=Undefined, new=self[-1])

    @_control_extend.before
    def _control_extend(self, call, notify):
        return len(self)

    @_control_extend.after
    def _control_extend(self, answer, notify):
        for i in range(answer.before, len(self)):
            notify(index=i, old=Undefined, new=self[i])

    @_control_pop.after
    def _control_pop(self, answer, notify):
        notify(index=len(self), old=answer.value, new=Undefined)

    @_control_clear.before
    def _control_clear(self, call, notify):
        return self.copy()

    @_control_clear.after
    def _control_clear(self, answer, notify):
        for i, v in enumerate(reversed(answer.before)):
            notify(index=i, old=v, new=Undefined)

    @_control_remove.before
    def _control_remove(self, call, notify):
        index = self.index(call.args[0])
        return index, self[index:]

    _control_remove.after(_control_delitem)

    def _control_rearrangement(self):
        old = self[:]
        def _after_rearangement(returned, notify):
            for i, v in enumerate(old):
                if v != self[i]:
                    notify(index=i, old=v, new=self[i])
        return _after_rearangement


class Dict(Model, dict):
    """An MVC enabled ``dict``."""

    _model_selector_template = '{key}'

    _control_setitem = Control('__setitem__', 'setdefault')
    _control_delitem = Control('__delitem__', 'pop')
    _control_update = Control('update')
    _control_clear = Control('clear')

    @_control_setitem.before
    def _control_setitem(self, call, notify):
        key = call.args[0]
        old = self.get(key, Undefined)
        return key, old

    @_control_setitem.after
    def _control_setitem(self, answer, notify):
        key, old = answer.before
        new = self[key]
        if new != old:
            notify(key=key, old=old, new=new)

    @_control_delitem.before
    def _control_delitem(self, call, notify):
        key = call.args[0]
        try:
            old = self[key]
        except KeyError:
            pass
        else:
            def _after(returned):
                notify(key=key, old=old, new=Undefined)
            return _after

    @_control_update.before
    def _control_update(self, call, notify):
        if len(call.args):
            args = call.args[0]
            if inspect.isgenerator(args):
                # copy generator so it doesn't get exhausted
                args = itertools.tee(arsg)[1]
            new = dict(args)
            new.update(call.kwargs)
        else:
            new = call.kwargs
        old = {k: self.get(k, Undefined) for k in new}
        return old

    @_control_update.after
    def _control_update(self, answer, notify):
        for k, v in answer.before.items():
            if self[k] != v:
                notify(key=k, old=v, new=self[k])

    @_control_clear.before
    def _control_clear(self, call, notify):
        return self.copy()

    @_control_clear.after
    def _control_clear(self, answer, notify):
        for k, v in answer.before.items():
            notify(key=k, old=v, new=Undefined)


class Set(Model, set):
    """An MVC enabled ``set``."""

    _control_update = Control(
        "clear", "update", "difference_update", "intersection_update",
        "add", "pop", "remove", "symmetric_difference_update", "discard",
    )

    @_control_update.before
    def _control_update(self, call, notify):
        return self.copy()

    @_control_update.after
    def _control_update(self, answer, notify):
        new = self.difference(answer.before)
        old = answer.before.difference(self)
        if new or old:
            notify(new=new, old=old)


class Object(Model):

    _control_attr_change = Control('__setattr__', '__delattr__')

    @_control_attr_change.before
    def _control_attr_change(self, call, notify):
        return call.args[0], getattr(self, call.args[0], Undefined)

    @_control_attr_change.after
    def _control_attr_change(self, answer, notify):
        attr, old = answer.before
        new = getattr(self, attr, Undefined)
        if new != old:
            notify(attr=attr, old=old, new=new)


class Oracle(Object):

    def __init__(self, **attrs):

        def try_to_view(path, x):
            if isinstance(x, Dict):
                view(x)(MethodType(notify_dict_change, path))
                for k, v in x.items():
                    try_to_view(path + (k,), v)
            elif isinstance(x, List):
                view(x)(MethodType(notify_list_change, path))
                for i, v in enumerate(x):
                    try_to_view(path + (i,), v)
            elif isinstance(x, Object):
                view(x)(MethodType(notify_object_change, path))
                for k, v in x.__dict__.items():
                    try_to_view(path + (k,), v)
            elif isinstance(x, Set):
                view(x)(MethodType(notify_set_change, path))
                for v in x:
                    try_to_view(path, v)
            else:
                try:
                    hash(x)
                except TypeError:
                    name = type(self).__name__
                    msg = "%r cannot assign the mutable value %r which is not a 'Model' type"
                    raise TypeError(msg % (name, x))

        def notify_object_change(path, events, oracle=ref(self)):
            new_events = []
            for e in events:
                try_to_view(path + (e["attr"],), e["new"])
                new_events.append(e["location": path])
            oracle()._notify_oracle_views(new_events)

        def notify_dict_change(path, events, oracle=ref(self)):
            new_events = []
            for e in events:
                try_to_view(path + (e["key"],), e["new"])
                new_events.append(e["location": path])
            oracle()._notify_oracle_views(new_events)

        def notify_list_change(path, events, oracle=ref(self)):
            new_events = []
            for e in events:
                try_to_view(path + (e["index"],), e["new"])
                new_events.append(e["location": path])
            oracle()._notify_oracle_views(new_events)

        def notify_set_change(path, events):
            oracle()._notify_oracle_event([e["location": path] for e in events])

        object.__setattr__(self, "_notify_root_view", notify_object_change)
        for k, v in attrs.items():
            try_to_view((k,), v)
        self.__dict__.update(attrs)

    def _notify_oracle_views(self, events):
        events = tuple(events)
        for view in self._model_views:
            view(events)

    def _notify_model_views(self, events):
        self._notify_root_view((), events)


# The MIT License (MIT)

# Copyright (c) 2016 Ryan S. Morshead

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import sys
from pytest import importorskip, fixture, raises

# skips for python 3.6 or less
mvc = importorskip('spectate.mvc')
from spectate.spectate import MethodSpectator


def model_events(mtype, *args, **kwargs):
    events = []
    model = mtype(*args, **kwargs)

    @mvc.view(model)
    def cache(e):
        events.extend(e)

    return model, events


class TestControl:

    def test_control_is_only_for_model(self):
        """Controls must be defined on a Model"""
        with raises(RuntimeError):
            class X:
                _control = mvc.Control()

    def test_control_decoration(self):

        calls = []

        class X(mvc.Model):

            def method(self):
                pass

            control = mvc.Control('method')

            @control.before
            def control(self, call, notify):
                calls.append('before')

            @control.after
            def control(self, call, notify):
                calls.append('after')

        assert isinstance(X.method, MethodSpectator)

        X().method()
        assert calls == ['before', 'after']


    def test_control_string_reference(self):

        calls = []

        class X(mvc.Model):

            def method(self): pass

            control = mvc.Control('method').before('before').after('after')

            def before(self, call, notify):
                calls.append('before')

            def after(self, answer, notify):
                calls.append('after')

        X().method()
        assert calls == ['before', 'after']


class TestView:

    class Counter(mvc.Model):

        def __init__(self):
            self.x = 0

        def increment(self, amount):
            self.x += amount

        def decrement(self, amount):
            self.x -= amount

        # define a beforeback for increment and decrement
        _control_change = mvc.Control('increment', 'decrement')

        @_control_change.before
        def _control_change(self, call, notify):
            return self.x

        # create the corresponding afterback
        @_control_change.after
        def _control_change(self, answer, notify):
            # Send an "event" dictionary to the Counter's views.
            notify(old=answer.before, new=self.x)

    def test_custom(self):
        counter, events = model_events(self.Counter)

        counter.increment(1)
        counter.decrement(2)
        counter.increment(3)
        counter.decrement(4)

        assert events == [
            {'old': 0, 'new': 1},
            {'old': 1, 'new': -1},
            {'old': -1, 'new': 2},
            {'old': 2, 'new': -2},
        ]

    def test_hold(self):
        counter, events = model_events(self.Counter)

        with mvc.hold(counter) as cache:
            counter.increment(1)
            assert cache == [
                {'old': 0, 'new': 1},
            ]

            counter.increment(1)
            assert cache == [
                {'old': 0, 'new': 1},
                {'old': 1, 'new': 2},
            ]

            # Pop off one of the events so
            # it isn't sent to notifiers.
            cache.pop()

        assert events == [
            {'old': 0, 'new': 1},
        ]

    def test_mute(self):
        counter, events = model_events(self.Counter)

        with mvc.mute(counter) as cache:
            counter.increment(1)
            counter.increment(1)

        assert events == []


def test_memory_safety():

    l = mvc.List()

    @mvc.view(l)
    def closure(event):
        # adds `l` to __closure__
        x = l

    @mvc.view(l)
    def defaults(event, l=l):
        # adds `l` to defaults
        x = l

    # Both of the above callbacks would normally create cirular
    # references, however mvc.utils.memory_safe_function will
    # mitigate this magically for the user.

    assert sys.getrefcount(l) == 2



# See End Of File For Licensing

import re
import sys
import six
import types
import inspect
import collections

try:
    from inspect import signature, Parameter, Signature
except ImportError:
    from funcsigs import signature, Parameter, Signature

__all__ = [
    'expose',
    'expose_as',
    'watch',
    'watched',
    'unwatch',
    'watcher',
    'watchable',
    'Watchable',
    'Data',
]


def _safe_signature(func):
    try:
        return signature(func)
    except ValueError:
        # builtin methods don't have sigantures
        return Signature([
            Parameter("args", Parameter.VAR_POSITIONAL),
            Parameter("kwargs", Parameter.VAR_KEYWORD),
        ])


def _signature_breakdown(func):
    args = []
    defaults = []
    var_keyword = None
    var_positional = None
    sig = _safe_signature(func)
    for param in sig.parameters.values():
        if param.kind == Parameter.VAR_POSITIONAL:
            var_positional = param.name
        elif param.kind == Parameter.VAR_KEYWORD:
            var_keyword = param.name
        else:
            if param.default is not Parameter.empty:
                defaults.append(param.default)
            args.append(param.name)
    return str(sig), tuple(args), defaults, var_positional, var_keyword


class Spectator(object):
    """An object for holding callbacks"""

    def __init__(self, subclass):
        """Create a Spectator that can be registered to a :class:`Watchable` instance.

        Parameters
        ----------
        subclass: type
            A the :class:`Watchable` subclass whose instance this :class:`Specatator` can respond to.
        """
        if not issubclass(subclass, Watchable):
            raise TypeError('Expected a Watchable, not %r.' % subclass)
        self.subclass = subclass
        self._callback_registry = {}

    def callback(self, name, before=None, after=None):
        """Add a callback pair to this spectator.

        You can specify, with keywords, whether each callback should be triggered
        before, and/or or after a given method is called - hereafter refered to as
        "beforebacks" and "afterbacks" respectively.

        Parameters
        ----------
        name: str
            The name of the method to which callbacks should respond.
        before: None or callable
            A callable of the form ``before(obj, call)`` where ``obj`` is
            the instance which called a watched method, and ``call`` is a
            :class:`Data` containing the name of the called method, along with
            its positional and keyword arguments under the attributes "name"
            "args", and "kwargs" respectively.
        after: None or callable
            A callable of the form ``after(obj, answer)`` where ``obj` is
            the instance which alled a watched method, and ``answer`` is a
            :class:`Data` containing the name of the called method, along with
            the value it returned, and data ``before`` may have returned
            under the attributes "name", "value", and "before" respectively.
        """
        if isinstance(name, (list, tuple)):
            for name in name:
                self.callback(name, before, after)
        else:
            if not isinstance(getattr(self.subclass, name), MethodSpectator):
                raise ValueError("No method specator for '%s'" % name)
            if before is None and after is None:
                raise ValueError("No pre or post '%s' callbacks were given" % name)
            elif before is not None and not callable(before):
                raise ValueError("Expected a callable, not %r." % before)
            elif after is not None and not callable(after):
                raise ValueError("Expected a callable, not %r." % after)
            elif before is None and after is None:
                raise ValueError("No callbacks were given.")
            if name in self._callback_registry:
                l = self._callback_registry[name]
            else:
                l = []
                self._callback_registry[name] = l
            l.append((before, after))

    def remove_callback(self, name, before=None, after=None):
        """Remove a beforeback, and afterback pair from this Spectator

        If ``before`` and ``after`` are None then all callbacks for
        the given method will be removed. Otherwise, only the exact
        callback pair will be removed.

        Parameters
        ----------
        name: str
            The name of the method the callback pair is associated with.
        before: None or callable
            The beforeback that was originally registered to the given method.
        after: None or callable
            The afterback that was originally registered to the given method.
        """
        if isinstance(name, (list, tuple)):
            for name in name:
                self.remove_callback(name, before, after)
        elif before is None and after is None:
            del self._callback_registry[name]
        else:
            if name in self._callback_registry:
                l = self._callback_registry[name]
            else:
                l = []
                self._callback_registry[name] = l
            l.remove((before, after))
            if len(l) == 0:
                # cleanup if all callbacks are gone
                del self._callback_registry[name]

    def wrapper(self, name, args, kwargs):
        """Trigger a method along with its beforebacks and afterbacks.

        Parameters
        ----------
        name: str
            The name of the method that will be called
        args: tuple
            The arguments that will be passed to the base method
        kwargs: dict
            The keyword args that will be passed to the base method
        """
        ms = getattr(self.subclass, name)
        if not isinstance(ms, MethodSpectator):
            raise TypeError(
                "'%s.%s' has no MethodSpectator" % (
                self.subclass.__name__, name))
        if name in self._callback_registry:
            beforebacks, afterbacks = zip(*self._callback_registry.get(name, []))

            hold = []
            for b in beforebacks:
                if b is not None:
                    call = Data(name=name,
                        kwargs=kwargs.copy(),
                        args=args[1:])
                    v = b(args[0], call)
                else:
                    v = None
                hold.append(v)

            out = ms.basemethod(*args, **kwargs)

            for a, bval in zip(afterbacks, hold):
                if a is not None:
                    a(args[0], Data(before=bval,
                        name=name, value=out))
                elif callable(bval):
                    # the beforeback's return value was an
                    # afterback that expects to be called
                    bval(out)
            return out
        else:
            return ms.basemethod(*args, **kwargs)


class MethodSpectator(object):
    """Notifies a :class:`Specator` when the method this descriptor wraps is called."""

    _compile_count = 0
    _src_str = """def {name}({signature}):
    args, vargs, kwargs = {arguments}, {variable_positional}, {variable_keyword};
    return globals()["spectator"].wrapper('{name}', (args + vargs), kwargs)"""

    def __init__(self, basemethod, name=None):
        if not callable(basemethod):
            raise TypeError('Expected a callable, not %r' % basemethod)
        self.basemethod = basemethod
        self.name = name or basemethod.__name__
        self.code, self.defaults = self._code(basemethod)

    def _code(self, basemethod):
        sig, args, defaults, varargs, varkwargs = _signature_breakdown(basemethod)
        src = self._src_str.format(
            name=self.name,
            # signature can end with ", /" for positional only params
            signature=str(sig)[1:-1].replace(", /", ""),
            arguments=str(tuple(args)).replace("'", ""),
            variable_positional=varargs or (),
            variable_keyword=varkwargs or {},
        )
        name = re.findall('[A-Z][a-z]*', type(self).__name__)
        filename = "-".join(name).upper() + "-#%s"
        code = compile(src, filename % self._compile_count, 'single')
        type(self)._compile_count += 1
        return code, tuple(defaults)

    def new_wrapper(self, inst, spectator):
        evaldict = {"spectator": spectator}
        eval(self.code, evaldict)
        # extract wrapper by name
        new = evaldict[self.name]
        # assign docstring and defaults
        new.__doc__ = self.basemethod.__doc__
        new.__defaults__ = self.defaults
        return types.MethodType(new, inst)

    def __get__(self, inst, cls):
        if inst is None:
            return self
        elif getattr(inst, "_instance_spectator", None):
            return self.new_wrapper(inst, inst._instance_spectator)
        else:
            return types.MethodType(self.basemethod, inst)


class Watchable(object):
    """A base class for introspection.

    And in Python>=3.6 rewraps overriden methods with a :class:`MethodSpectator`
    if appropriate.
    """

    if not sys.version_info < (3, 6):

        def __init_subclass__(cls, **kwargs):
            """If a subclass overrides a :class:`MethodSpectator` method, then rewrap it."""
            for base in cls.mro()[1:]:
                if issubclass(base, Watchable):
                    for k, v in base.__dict__.items():
                        if k in cls.__dict__ and isinstance(v, MethodSpectator):
                            new = getattr(cls, k)
                            if callable(new) and not isinstance(new, MethodSpectator):
                                method_spectator = MethodSpectator(new, k)
                                setattr(cls, k, method_spectator)
            super().__init_subclass__(**kwargs)


def expose(*methods):
    """A decorator for exposing the methods of a class.

    Parameters
    ----------
    *methods : str
        A str representation of the methods that should be exposed to callbacks.

    Returns
    -------
    decorator : function
        A function accepting one argument - the class whose methods will be
        exposed - and which returns a new :class:`Watchable` that will
        notify a :class:`Spectator` when those methods are called.

    Notes
    -----
    This is essentially a decorator version of :func:`expose_as`
    """
    def setup(base):
        return expose_as(base.__name__, base, *methods)
    return setup


def expose_as(name, base, *methods):
    """Return a new type with certain methods that are exposed to callback registration.

    Parameters
    ----------
    name : str
        The name of the new type.
    base : type
        A type such as list or dict.
    *methods : str
        A str representation of the methods that should be exposed to callbacks.

    Returns
    -------
    exposed : obj:
        A :class:`Watchable` with methods that will notify a :class:`Spectator`.
    """
    classdict = {}
    for method in methods:
        if not hasattr(base, method):
            raise AttributeError("Cannot expose '%s', because '%s' "
                "instances lack this method" % (method, base.__name__))
        else:
            classdict[method] = MethodSpectator(getattr(base, method), method)
    return type(name, (base, Watchable), classdict)


def watchable(value):
    """Returns True if the given value is a subclass or instance of :class:`Watchable`."""
    check = issubclass if inspect.isclass(value) else isinstance
    return check(value, Watchable)


def watch(value, spectator_type=Spectator):
    """Register a :class:`Specatator` to a :class:`Watchable` and return it.

    In order to register callbacks to an eventful object, you need to create
    a Spectator that will watch it for you. A :class:`Specatator` is a relatively simple
    object that has methods for adding, deleting, and triggering callbacks. To
    create a spectator we call ``spectator = watch(x)``, where x is a Watchable
    instance.

    Parameters
    ----------
    value : Watchable
        A :class:`Watchable` instance.
    spectator_type : Spectator
        The type of spectator that will be returned.

    Returns
    -------
    spectator: spectator_type
        The :class:`Specatator` (specified by ``spectator_type``) that is
        was registered to the given instance.
    """
    if isinstance(value, Watchable):
        wtype = type(value)
    else:
        raise TypeError("Expected a Watchable, not %r." % value)
    spectator = getattr(value, "_instance_spectator", None)
    if not isinstance(spectator, Spectator):
        spectator = spectator_type(wtype)
        value._instance_spectator = spectator
    return spectator


def watched(cls, *args, **kwargs):
    """Create and return a :class:`Watchable` with its :class:`Specatator`.

    See :func:`watch` for more info on :class:`Specatator` registration.

    Parameters
    ----------
    cls: type:
        A subclass of :class:`Watchable`
    *args:
        Positional arguments used to create the instance
    **kwargs:
        Keyword arguments used to create the instance.
    """
    value = cls(*args, **kwargs)
    return value, watch(value)


def unwatch(value):
    """Return the :class:`Specatator` of a :class:`Watchable` instance."""
    if not isinstance(value, Watchable):
        raise TypeError("Expected a Watchable, not %r." % value)
    spectator = watcher(value)
    try:
        del value._instance_spectator
    except:
        pass
    return spectator


def watcher(value):
    """Return the :class:`Specatator` of a :class:`Watchable` instance."""
    if not isinstance(value, Watchable):
        raise TypeError("Expected a Watchable, not %r." % value)
    return getattr(value, "_instance_spectator", None)


class Data(collections.Mapping):
    """An immutable mapping with attribute-access.

    Empty keys are represented with a value of ``None``.

    In order to evolve :class:`Data`, users must create copies that
    contain updates:

    .. code-block:: python

        d1 = Data(a=1)
        d2 = Data(b=2)
        assert Data(d1, **d2) == {'a': 1, 'b': 2}

    Easing this fact, is :class:`Data`'s syntactic sugar:

    .. code-block:: python

        d1 = Data(a=1)
        assert d1 == {'a': 1}

        d2 = d1['b': 2]
        assert d2 == {'a': 1, 'b': 2}

        d3 = d2['a': None, 'b': 1]
        assert d3 == {'b': 1}

        d4 = d3[{'a': 1, 'c': 3}, {'b': None}]
        assert d4 == {'a': 1, 'c': 3}
    """

    def __init__(self, *args, **kwargs):
        items = dict(*args, **kwargs).items()
        self.__dict__.update(i for i in items if i[1] is not None)

    def __getattr__(self, key):
        return None

    def __getitem__(self, key):
        if type(key) is slice:
            key = (key,)
        if type(key) is tuple:
            for x in key:
                if not isinstance(x, slice):
                    break
            else:
                new = {s.start : s.stop for s in key}
                return type(self)(self, **new)
            merge = {}
            for x in key:
                if isinstance(x, collections.Mapping):
                    merge.update(x)
            key = merge
        if isinstance(key, collections.Mapping):
            return type(self)(self, **key)
        return self.__dict__.get(key)

    def __setitem__(self, key, value):
        raise TypeError('%r is immutable')

    def __setattr__(self, key, value):
        raise TypeError('%r is immutable')

    def __delitem__(self, key):
        raise TypeError('%r is immutable')

    def __delattr__(self, key):
        raise TypeError('%r is immutable')

    def __contains__(self, key):
        return key in tuple(self)

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        return repr(self.__dict__)


# The MIT License (MIT)

# Copyright (c) 2016 Ryan S. Morshead

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



"""
MVC
===

.. automodule:: spectate.mvc.base
    :members:
    :ignore-module-all:

.. automodule:: spectate.mvc.models
    :members:
    :ignore-module-all:

.. automodule:: spectate.mvc.utils
    :members:
"""

import sys

if sys.version_info < (3, 6):
    raise ImportError('Python 3.6 or greater required.')
else:
    del sys

from .base import *
from .models import *



# See End Of File For Licensing

from contextlib import contextmanager
from functools import wraps, partial
from collections import defaultdict

from .utils import memory_safe_function, Sentinel
from ..spectate import Watchable, Data, MethodSpectator, expose, watched

__all__ = [
    'Model',
    'Control',
    'view',
    'unview',
    'is_model',
    'hold',
    'mute',
]


def is_model(model, raises=False):
    if isinstance(model, Model):
        return True
    elif raises:
        raise TypeError('Expected a Model, not %r.' % model)
    else:
        return False


@contextmanager
def hold(model):
    is_model(model, raises=True)
    events = []
    redirect = lambda e : events.extend(e)
    restore = model.__dict__.get('_notify_model_views')
    model._notify_model_views = redirect

    try:
        yield events
    finally:
        if restore is None:
            del model._notify_model_views
        else:
            model._notify_model_views = restore
        model._notify_model_views(tuple(events))


@contextmanager
def mute(model):
    is_model(model, raises=True)
    redirect = lambda e : None
    restore = model.__dict__.get('_notify_model_views')
    model._notify_model_views = redirect
    try:
        yield
    finally:
        if restore is None:
            del model._notify_model_views
        else:
            model._notify_model_views = restore


def views(model):
    """Return a model's views keyed on what events they respond to."""
    is_model(model, raises=True)
    return model._model_views[:]


def view(model):
    is_model(model, raises=True)
    def setup(function):
        # we want to avoid circular references.
        safe = memory_safe_function(function)
        model._model_views.append(safe)
        return safe
    return setup


def unview(model, function):
    model._model_views.remove(function)


class Control:

    def __init__(self, *methods):
        self.methods = methods
        self.name = None

    def before(self, callback):
        if isinstance(callback, Control):
            callback = callback._before
        elif not isinstance(callback, str):
            callback = self._wrap(callback)
        self._before = callback
        return self

    def after(self, callback):
        if isinstance(callback, Control):
            callback = callback._after
        elif not isinstance(callback, str):
            callback = self._wrap(callback)
        self._after = callback
        return self

    @staticmethod
    def _wrap(function):
        if function is None:
            return None
        @wraps(function)
        def callback(self, *args, **kwargs):
            events = []
            def notify(**event):
                events.append(Data(event))
            args = args + (notify,)
            result = function(self, *args, **kwargs)
            if events:
                self._notify_model_views(tuple(events))
            return result
        return callback

    _after, _before = None, None

    def __set_name__(self, cls, name):
        if not issubclass(cls, Model):
            raise TypeError("Can only define a control on a Model, not %r" % cls)
        if self.name:
            msg = 'Control was defined twice - %r and %r.'
            raise RuntimeError(msg % (self.name, name))
        else:
            self.name = name
        if isinstance(self._after, str):
            self._after = self._wrap(getattr(cls, self._after))
        if isinstance(self._before, str):
            self._before = self._wrap(getattr(cls, self._before))
        for m in self.methods:
            setattr(cls, m, MethodSpectator(getattr(cls, m), m))


class Model(Watchable):

    _model_controls = ()

    def __init_subclass__(cls, **kwargs):
        for k, v in list(cls.__dict__.items()):
            if isinstance(v, Control):
                cls._model_controls += (k,)
        super().__init_subclass__(**kwargs)

    def __new__(cls, *args, **kwargs):
        self, spectator = watched(super().__new__, cls)
        for name in cls._model_controls:
            ctrl = getattr(cls, name)
            for method in ctrl.methods:
                before, after = ctrl._before, ctrl._after
                spectator.callback(method, before, after)
        object.__setattr__(self, "_model_views", [])
        return self

    def _notify_model_views(self, events):
        for view in self._model_views:
            view(events)


_Empty = Sentinel('Empty')

# The MIT License (MIT)

# Copyright (c) 2016 Ryan S. Morshead

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from spectate import expose, watch


@expose('__setattr__', '__delattr__')
class Data(object):

    def __init__(self):
        self._callbacks = []
        spectator = watch(self)
        spectator.callback(
            '__setattr__',
            self._before_update,
            self._after_update,
        )
        spectator.callback(
            '__delattr__',
            self._before_update,
            self._after_update,
        )

    @staticmethod
    def _before_update(data, call):
        key = call.args[0]
        old = data.__dict__.get(key)
        return key, old

    @staticmethod
    def _after_update(data, answer):
        key, old = answer.before
        new = data.__dict__.get(key)
        for c in data._callbacks:
            c(key=key, old=old, new=new)

    def watch(self, function):
        self._callbacks.append(function)
        return function


class User(Data):

    def __init__(self, name, description):
        super(User, self).__init__()
        self.name = name
        self.description = description


user = User('Paul', 'a new user')


@user.watch
def printer(key, old, new):
    print('%s : %r -> %r' % (key, old, new))


user.name = 'Paul Jones'
user.description = 'updated user description'



from spectate import expose, watch


@expose('increment', 'decrement')
class Counter(object):

    def __init__(self):
        self.x = 0

    def increment(self, amount):
        self.x += amount

    def decrement(self, amount):
        self.x -= amount


counter = Counter()
spectator = watch(counter)


def changed(counter, answer):
    print(counter.x)


spectator.callback('increment', after=changed)
spectator.callback('decrement', after=changed)

# see what happens when we modify the counter:

counter.increment(1)
counter.decrement(2)
counter.increment(3)
counter.decrement(4)



"""
Spectate
========

.. automodule:: spectate.spectate
    :members:
    :ignore-module-all:

.. automodule:: spectate.mvc
    :members:
"""

from .spectate import *
from ._version import __version__



import sys
import inspect
from pytest import raises
from spectate.spectate import (
    expose, expose_as, watch, watched,
    watcher, unwatch, watchable, Watchable,
    MethodSpectator, Spectator, Data,
)

try:
    from inspect import signature
except ImportError:
    from funcsigs import signature


def test_watchable():
    assert watchable(Watchable)
    assert watchable(Watchable())


def test_expose():

    @expose('increment')
    class Counter(object):

        def __init__(self):
            self.x = 0

        def increment(self, amount=1):
            self.x += amount

    assert watchable(Counter)
    assert Counter.__name__ == 'Counter'
    assert isinstance(Counter.increment, MethodSpectator)


def test_expose_as():
    WatchableList = expose_as("WatchableList", list, 'append')
    assert watchable(WatchableList)
    assert issubclass(WatchableList, list)
    assert WatchableList.__name__ == 'WatchableList'
    assert isinstance(WatchableList.append, MethodSpectator)


class Thing(object):

    def func(self, a, b, c=None, d=None, *e, **f):
        return (self, a, b, c, d, e, f)


def test_watch():
    WatchableThing = expose_as("WatchableThing", Thing, "func")
    wt = WatchableThing()
    spectator = watch(wt)

    assert isinstance(spectator, Spectator)
    assert hasattr(wt, "_instance_spectator")
    assert wt._instance_spectator is spectator

    WatchableList = expose_as("WatchableList", list, "append")
    wl, spectator = watched(WatchableList, [1, 2, 3])

    assert wl == [1, 2, 3]
    assert isinstance(spectator, Spectator)
    assert hasattr(wl, "_instance_spectator")
    assert wl._instance_spectator is spectator


def test_watched():
    WatchableThing = expose_as("WatchableThing", Thing, "func")
    wt, spectator = watched(WatchableThing)

    assert isinstance(spectator, Spectator)
    assert hasattr(wt, "_instance_spectator")
    assert wt._instance_spectator is spectator

    # Since watched uses watch under the hood
    # we don't need the remainder of the tests.


def test_watcher():
    WatchableThing = expose_as("WatchableThing", Thing, "func")
    wt, spectator = watched(WatchableThing)
    assert watcher(wt) is spectator


def test_unwatch():
    WatchableThing = expose_as("WatchableThing", Thing, "func")
    wt, spectator = watched(WatchableThing)
    out = unwatch(wt)
    assert not hasattr(wt, "_instance_spectator")
    assert out is spectator


def test_method_spectator():
    MethodSpectator._compile_count = 0
    WatchableList = expose_as("WatchableList", list, 'append')
    assert MethodSpectator._compile_count == 1
    append = WatchableList.append

    assert append.basemethod is list.append
    assert append.name == 'append'

    wl, spectator = watched(WatchableList)
    wl.append(1)
    wl.append(2)
    assert wl == [1, 2]


def test_method_spectator_signature():
    WatchableThing = expose_as("WatchableThing", Thing, 'func')
    thing, sectator = watched(WatchableThing)
    assert MethodSpectator._compile_count == 2
    assert (
        signature(Thing().func) == signature(thing.func)
    )


def check_answer(checklist, inst, name, a, b, c=None, d=None, *e, **f):
    args, kwargs = condense(a, b, c, d, *e, **f)
    checklist.append(Data(
        name=name,
        value=(inst, a, b, c, d, e, f),
        before=Data(
            name=name,
            args=args,
            kwargs=kwargs))
    )
    getattr(inst, name)(a, b, c, d, *e, **f)


condense = lambda *a, **kw: (a, kw)


def test_beforeback_afterback():
    checklist = []

    WatchableThing = expose_as("WatchableThing", Thing, 'func')
    wt = WatchableThing()
    spectator = watch(wt)

    callbacks_called = [0, 0]

    # callback stores call information
    def beforeback(inst, call):
        callbacks_called[0] += 1
        return call
    def afterback(inst, answer):
        callbacks_called[1] += 1
        assert checklist[-1] == answer

    spectator.callback('func',
        before=beforeback, after=afterback)

    check_answer(checklist, wt, 'func', 1, 2, c=3)
    check_answer(checklist, wt, 'func', 1, 2, d=3)
    check_answer(checklist, wt, 'func', 1, 2, 3, 4, 5)
    check_answer(checklist, wt, 'func', 1, 2, d=3, f=4)
    assert callbacks_called == [4, 4]


def test_callback_closure():
    checklist = []

    WatchableThing = expose_as("WatchableThing", Thing, 'func')
    wt = WatchableThing()
    spectator = watch(wt)

    callbacks_called = [0, 0]

    def callback(inst, call):
        callbacks_called[0] += 1
        def closure(value):
            callbacks_called[1] += 1
            assert (checklist[-1] == Data(
                name=call.name, value=value,
                before=call))
        return closure

    spectator.callback('func', callback)

    check_answer(checklist, wt, 'func', 1, 2, c=3)
    check_answer(checklist, wt, 'func', 1, 2, d=3)
    check_answer(checklist, wt, 'func', 1, 2, 3, 4, 5)
    check_answer(checklist, wt, 'func', 1, 2, d=3, f=4)
    assert callbacks_called == [4, 4]


def test_callback_multiple():

    class Test(object):

        def a(self):
            pass

        def b(self):
            pass

    WatchableTest = expose_as("WatchableTest", Test, "a", "b")
    wt, spectator = watched(WatchableTest)

    callback = lambda value, call: None

    spectator.callback(("a", "b"), callback)

    for key in ("a", "b"):
        assert key in spectator._callback_registry
        assert spectator._callback_registry[key] == [(callback, None)]

    spectator.remove_callback(("a", "b"), callback)

    assert spectator._callback_registry == {}


if not sys.version_info < (3, 6):

    def test_subclass_override():

        @expose('method')
        class Parent:

            def method(self):
                pass

        class Child(Parent):

            def method(self):
                pass

        assert watchable(Child)
        assert isinstance(Child.method, MethodSpectator)


def test_data_is_immutable():
    d = Data(a=0)
    with raises(TypeError):
        d['a'] = 1
    with raises(TypeError):
        d.a = 1
    with raises(TypeError):
        del d['a']
    with raises(TypeError):
        del d.a
    assert d == {'a': 0}


def test_none_is_empty_data():
    d = Data(a=None)
    assert d == {}
    assert 'b' not in d
    assert d['a'] is None


def test_data_evolution():
    d0 = Data(a=0)
    d1 = d0['b': 1]
    assert d1 == {'a': 0, 'b': 1}
    d2 = d1['b': 2, 'c': 3]
    assert d2 == {'a': 0, 'b': 2, 'c': 3}
    d3 = d2[{'b': 3, 'c': 4}]
    d4 = d3['b': None, 'c': None]
    assert d4 == d0
    d5 = d4[{'a': None}]
    assert d5 == {}
    d6 = d5[{'a': 1}, {'b': 2}]
    assert d6 == {'a': 1, 'b': 2}

def test_data_is_mapping():
    assert dict(Data(a=0, b=1)) == {'a': 0, 'b': 1}
    assert dict(**Data(a=0, b=1)) == {'a': 0, 'b': 1}



from __future__ import print_function
from setuptools import find_packages

# the name of the project
name = "spectate"

#-----------------------------------------------------------------------------
# Minimal Python version sanity check
#-----------------------------------------------------------------------------

import sys

v = sys.version_info
if v[:2] < (2,7) or (v[0] >= 3 and v[:2] < (3,3)):
    error = "ERROR: %s requires Python version 2.7 or 3.3 or above." % name
    print(error, file=sys.stderr)
    sys.exit(1)

#-----------------------------------------------------------------------------
# get on with it
#-----------------------------------------------------------------------------

import os
from glob import glob

from distutils.core import setup

here = os.path.abspath(os.path.dirname(__file__))
root = os.path.join(here, name)

packages = find_packages()


requirements = [
    'six',
]


with open(os.path.join(root, '_version.py')) as f:
    namespace = {}
    exec(f.read(), {}, namespace)
    version = namespace["__version__"]


with open(os.path.join(here, 'README.md')) as f:
    long_description = f.read()


if __name__ == '__main__':
    setup(
        name=name,
        version=version,
        packages=packages,
        description="Create classes whose instances have tracked methods",
        long_description=long_description,
        long_description_content_type='text/markdown',
        author="Ryan Morshead",
        author_email="ryan.morshead@gmail.com",
        url="https://github.com/rmorshea/spectate",
        license='MIT',
        platforms="Linux, Mac OS X, Windows",
        keywords=["eventful", "callbacks"],
        install_requires=requirements,
        classifiers=[
            'Intended Audience :: Developers',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.3',
            ],
    )



from spectate import mvc


class Counter(mvc.Model):

    def __init__(self):
        self.x = 0

    def increment(self, amount):
        self.x += amount

    def decrement(self, amount):
        self.x -= amount

    # define a control for incrementing and decrementing
    _control_change = mvc.Control('increment', 'decrement')

    # register a beforeback to the control
    @_control_change.before
    def _control_change(self, call, notify):
        return self.x

    # register an afterback to the control
    @_control_change.after
    def _control_change(self, answer, notify):
        # Send an "event" dictionary to the Counter's views.
        notify(old=answer.before, new=self.x)


counter = Counter()


@mvc.view(counter)
def printer(event):
    print(event)


counter.increment(1)
counter.decrement(2)
counter.increment(3)
counter.decrement(4)



import types
from weakref import proxy


class Sentinel:

    def __init__(self, name):
        self.__name = name

    def __repr__(self):
        return self.__name


def memory_safe_function(function):
    """Return a memory safe copy of a function.

    Through some clever tricks, all the defauts, and closure of this function
    are turned into proxy objects wherever possible. **Making a function memory
    safe requires its parent scope to be modified** - if the function contains
    a ``list``, ``dict``, or ``set`` in its closure or default arguments, all
    values contained in those data structures will be converted to proxy objects!
    Thus the user must keep references to those objects alive outside those data
    structures.

    Parameters
    ----------
    function: FunctionType, MethodType
        The function to be made into a memory safe copy.
    """
    if isinstance(function, types.MethodType):
        self = as_proxy(function.__self__)
        function = function.__func__
    elif not isinstance(function, types.FunctionType):
        raise TypeError('Expected a function or method, not %r.' % function)
    else:
        self = None

    closure = to_closure(as_proxy(cell.cell_contents) for cell in function.__closure__ or ())
    defaults = tuple(map(as_proxy, function.__defaults__ or ()))

    safe = types.FunctionType(
        function.__code__,
        function.__globals__,
        function.__name__,
        defaults,
        closure)

    if self is not None:
        safe = types.MethodType(safe, self)
    return safe


def as_proxy(x):
    if type(x) is list:
        for i, v in enumerate(x):
            x[i] = as_proxy(v)
        return x
    elif type(x) is dict:
        for k, v in x.items():
            x[as_proxy(k)] = as_proxy(v)
        return x
    elif type(x) is set:
        for v in x:
            x.remove(v)
            x.add(as_proxy(v))
        return x
    elif type(x) is tuple:
        return tuple(map(as_proxy, x))
    else:
        try:
            return proxy(x)
        except TypeError:
            return x


def to_closure(args):
    args = list(args)
    if not args:
        return ()
    numbers = tuple(range(len(args)))
    outer = '\n    '.join(map('out{0} = var{0}'.format, numbers))
    inner = '\n        '.join(map('in{0} = out{0}'.format, numbers))
    form = 'def outer():\n    {0}\n    def inner():\n        {1}\n    return inner'

    variables = {'var%s' % i : x for i, x in enumerate(args)}
    source = form.format(outer, inner)

    exec(source, variables)

    return variables['outer']().__closure__


class _dict(dict):
    """Can be weakref"""
